// Image filtering GPU processing lab
// by Ingemar Ragnemalm 2010

// Revised version 100311:
// Bug fix in readppm for non-square textures
// Some FBO code moved to "reusable"
// Shader loading now includes reading the file
// Fixed a bug in infolog printout

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef __APPLE__
	#include <OpenGL/gl.h>
	#include <GLUT/glut.h>
#else
	#include <GL/gl.h>
	#include <GL/glut.h>
#endif

#ifndef GL_VERSION_2_0
#error This program needs OpenGL 2.0 libraries to compile
#endif

#include "readppm.h"
#include "reusable.h"

// Globals

// Image dimensions
int n;
int m;

int lastw, lasth; // Needed to restore the viewport

// Reshape, set viewport and save it
void reshape(int h, int w)
{
	lastw = w;
	lasth = h;
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0.0, (GLfloat) n, 0.0, (GLfloat) m);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glViewport(0,0,n,m);
}


// Choose input data (textures) and output data (FBO)
void useFBO(struct FBOstruct *out, struct FBOstruct *in1, struct FBOstruct *in2)
{
	if (out != NULL && (out == in1 || out == in2))
		printf("Warning! Input and output identical!\n");
	
	if (out != 0L)
		glViewport(0, 0, out->width, out->height);
	else
		glViewport(0, 0, lastw, lasth);
	
	if (out != 0L)
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, out->fb);
	else
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glActiveTexture(GL_TEXTURE1);
	if (in2 != 0L)
		glBindTexture(GL_TEXTURE_2D, in2->texid);
	else
		glBindTexture(GL_TEXTURE_2D, 0);
	glActiveTexture(GL_TEXTURE0);
	if (in1 != 0L)
		glBindTexture(GL_TEXTURE_2D, in1->texid);
	else
		glBindTexture(GL_TEXTURE_2D, 0);
}

// Draw a single quad using the selected shader
void applyFilter(GLuint shaderProgramObject)
{
    // Erase in case we need it
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
	// run the edge detection filter over the geometry texture
	// Activate the edge detection filter program
	glUseProgram(shaderProgramObject);
	
	// Viewport-Sized Quad = Data Stream Generator.
	glBegin(GL_QUADS);
		glTexCoord2f(0, 0);
		glVertex2i(0, 0);
		glTexCoord2f(0, 1);
		glVertex2i(0, m);
		glTexCoord2f(1, 1);
		glVertex2i(n ,m);
		glTexCoord2f(1, 0);
		glVertex2i(n, 0);
	glEnd();
	
	// disable the filter
	glUseProgramObjectARB(0);
}

// ---------------------------------------------------------------------

// Set the texture unit uniforms for a shader.
// For this application, these can be the same in all shaders.
void setUniforms(GLuint s)
{
    GLint loc;
    
    glUseProgram(s);
    loc = glGetUniformLocation(s, "texUnit");
    glUniform1i(loc, 0); // texture unit 0
    loc = glGetUniformLocation(s, "texUnit2");
    glUniform1i(loc, 1); // texture unit 1
    glUseProgram(0);
}

// Application variables

unsigned char *image;
struct FBOstruct *fbo1, *fbo2, *fbo3;
GLuint	shaderProgramObject;	// the program used to update


void display()
{
	useFBO(NULL, fbo1, NULL); // Render to screen using fbo1 (input image) as input
	applyFilter(shaderProgramObject);

	glFlush();

// Other examples:	
// Render to screen without shader:
//	useFBO(NULL, fbo1, NULL); // Render to screen using fbo1 (input image) as input
//	applyFilter(0); // No shader (just copy the image from fbo1)
// Render from one fbo to another with shader:
//	useFBO(fbo2, fbo1, NULL); // Render to fbo2 using fbo1 as input
//	applyFilter(shaderProgramObject);
}

void keydown(unsigned char key, int x, int y)
{
	int redbits;

	switch (key)
	{
		case 'w':
		case 'W':
// Write result to disk
			glGetIntegerv(GL_RED_BITS, &redbits);
			printf("red bits %d\n", redbits);
			glDisable(GL_DITHER);
			glReadPixels(0, 0, n, m, GL_RGB, GL_UNSIGNED_BYTE, image);
			writeppm("rpout.ppm", n, m, image);
			break;
	}
}

int main(int argc, char **argv)
{	
	char *vs;
	char *fs;

    glutInit(&argc, argv);
    
// Input image
	image = readppm("maskros512.ppm", &n, &m);
	
// Standard OpenGL stuff
    glutInitDisplayMode(GLUT_SINGLE | GLUT_RGB );
    glutInitWindowSize(n, m);
	glutInitWindowPosition(50,50);
    glutCreateWindow("image");
	
// Create an FBO for each texture
	fbo1 = initFBO(m, n, image);
	fbo2 = initFBO(m, n, NULL);
	fbo3 = initFBO(m, n, NULL);

// Read and compile shader program(s)
	shaderProgramObject = loadAndCompileShaders("img.vs", "img.fs");
	
// Pass texture units to the shader program
// This is necessary to be able to use more than one texture!
    setUniforms(shaderProgramObject);
	
	reshape(n, m);
	
	glEnable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	
	glutReshapeFunc(reshape);
	glutDisplayFunc(display);
	glutKeyboardFunc(keydown);
	glutMainLoop();
}
